#include <assert.h>
#include <stdlib.h>

#include "heap.h"

//heap-static.h
//#define HEAP_CAPACITY 1024 // doit être une puissance de 2

#define HEAP_CAPACITY 8 // doit être une puissance de 2

/* with -std=c99 error with -std=c11 OK
static_assert((HEAP_CAPACITY != 0) && !(HEAP_CAPACITY & (HEAP_CAPACITY-1)),
	      "HEAP_CAPACITY n'est pas une puissance de 2");
*/

struct heap {
  int t[HEAP_CAPACITY];  // tas
  size_t size;           // taille du tas (nombre de nœuds)
};


//HEAP_DETAILS_H
/* NB: les fonctions ci-dessous mettent en œuvre un tas "min", c'est à dire
 * dont les valeurs sont extraites en ordre croissant
 */

/* Index du fils gauche d'un nœud
 * h : tas
 * i : nœud
 * RETURN : l'index du fils gauche de i dans le tas h s'il existe, -1 sinon
 */
static int heap__left(struct heap const * h, int i);

/* Index du fils droit d'un nœud
 * h : tas
 * i : nœud
 * RETURN : l'index du fils droit de i dans le tas h s'il existe, -1 sinon
 */
static int heap__right(struct heap const * h, int i);

/* Index du parent d'un nœud
 * h : tas
 * i : nœud
 * RETURN : l'index du parent de i dans le tas h s'il existe, -1 sinon
 */
static int heap__parent(struct heap const * h, int i);

/* Permute deux éléments du tas
 * h : tas
 * i : premier nœud
 * j : second nœud
 * POST : permute les éléments i et j du tas h si i et j sont des indices
 *        valides
 * RETURN : 0 si la permutation a eu lieu, -1 si i ou j n'est pas un indice
 *          valide dans h
 */
static int heap__swap(struct heap * h, int i, int j);

/* Remontre un nœud à sa place dans le tas
 * h : tas
 * i : nœud
 * PRE : h est un tas, sauf éventuellement à l'indice i où la valeur peut être
 *       plus petite qu'en certains ancêtres de i
 * POST : h est un tas qui contient les mêmes valeurs qu'initialement, mais en
 *        bonne place (la valeur en i a été remontée jusqu'à ce que son nœud
 *        parent ait une valeur plus petite, ou jusqu'à la racine)
 */
static void heap__bubble_up(struct heap * h, int i);

/* Descend un nœud à sa place dans le tas
 * h : tas
 * i : nœud
 * PRE : les sous-arbres gauche et droit de h sont des tas
 * POST : h est un tas qui contient les mêmes valeurs qu'initiallement, mais en
 *        bonne place (la valeur i a été descendue jusqu'à ne pas être plus
 *        grande que ses fils gauche et droit, ou jusqu'à une feuille)
 */
static void heap__bubble_down(struct heap * h, int i);


struct heap * heap__empty();

size_t heap__size(struct heap const * h);


int heap__add(struct heap * h, int x);

int heap__pick(struct heap * h, int * x);

void heap__free(struct heap * h);

static int heap__left(struct heap const * h, int i) {
  if ( (i < 0) || (i > (h->size-2)/2) )
    return -1;
  return (2*i+1);
}


static int heap__right(struct heap const * h, int i) {
  if ( (i < 0) || (i > (h->size-3)/2) )
    return -1;
  return (2*i+2);
}


static int heap__parent(struct heap const * h, int i) {
  if ( (i <= 0) || (i >= h->size) )
    return -1;
  return ((i-1)/2);
}


static int heap__swap(struct heap * h, int i, int j) {
  if ( (i < 0) || (i >= h->size) || (j < 0) || (j >= h->size) )
    return -1;
  int x = h->t[i];
  h->t[i] = h->t[j];
  h->t[j] = x;
  return 0;
}


static void heap__bubble_up(struct heap * h, int i) {
  if ( (i < 0) || (i >= h->size) )
    return;

  int index = i, parent = heap__parent(h, index);
  while ( (parent != -1) && (h->t[index] < h->t[parent]) ) {
    heap__swap(h, parent, index);
    index = parent;
    parent = heap__parent(h, index);
  }
}


static void heap__bubble_down(struct heap * h, int i) {
  if ( (i < 0) || (i >= h->size) )
    return;

  int index = i;
  while (1) {
    int left = heap__left(h, index), right = heap__right(h, index);
    int smallest = index;

    if ( (left != -1) && (h->t[left] < h->t[smallest]) )
      smallest = left;
    if ( (right != -1) && (h->t[right] < h->t[smallest]) )
      smallest = right;

    if (smallest == index)
      break;

    heap__swap(h, index, smallest);
    index = smallest;
  }
}
